//
// Class definitions and Member function definitions of CALC.CPP
//
// Programmer :-- Gaurang Ramakant Khetan
//

class identifier_table	// manages the identifier table i.e. variable values
	{
 private:
	int tot_var;							// total number of variables until now
	char var_str[MAX_VAR][MAX_VAR_LEN+1]; 			// variable strings
	float value[MAX_VAR];							// values of variables

	signed int index_of(const char*);  	// returns array index of variable
	int new_entry(const char*);			// returns array index after
													// initializing a new variable
 public:
	identifier_table() {	tot_var = 0; } 	//constructor

	//program interface routines
	float value_of(const char*);				// return value of a variable
	void set_value(const char*, float);		// assign a new value to
														// a new or old variable
	};

class input_buffer
	/* conveys the input from the user to the program funcs.
		If the input is to be redirected, or if an improvement is aspired in
		the keying in of input from user, modify the routines in this class,
		keeping the program interface routines functionally equivalent       */
	{
 private:
	char* inp_ptr;		// ptr to memory containing input string (dynamic alloc)
	int cur_pos;      // current position while reading string thru mem funcs
	int last_pos;     // position of last string character

 public:
	input_buffer()		// constructor
		{
		inp_ptr = (char*) malloc(MAX_INPUT);	// new & delete giving problems!
		cur_pos = 0; last_pos = 0;
		}
	void destroy()	{ free(inp_ptr); }	// pure destructor giving problems!

	// take input and store it in memory
	int take_input();

	// program interface routines
	void take_back()
		{ cur_pos--; }
	char next_input()
		{ if (cur_pos <= last_pos) return inp_ptr[cur_pos++]; else return 0;}
	};

class lexical_analyser
	/* this class lexically analyses the input buffer and conveys
		the result to the parser 												*/
	{
 private:
	input_buffer inp_buf;	//lexical_analyser contains the input buffer
	float number_value;     				// used when communicating with
	char string_value[MAX_VAR_LEN+1];	// parser.

	void get_number();	// recognize number in input string
	void get_string();   // recognize string in input string

 public:
	//pass message on to input_buffer
	int take_input()	{ return inp_buf.take_input(); }
	void destroy() { inp_buf.destroy(); }

	// Program interface routines
	token_types get_token();	//main routine
	float read_number()
		{ return number_value; }
	char* read_string()
		{ return string_value; }
	};

class parser
	{    // THE MAIN PARSER //
  private:
	token_types last_token;		//helps cooperation betn member funcs
	int in_paranthesis;	//keeps track of number of para.. we are in
	int to_take;

	lexical_analyser lex;  // takes help from lexical_analyser
	identifier_table tab;  // and identifier table

	float primary();	// takes care of bracket and unary minus
	float term();     // takes care of mult and div

  public:
	parser() { in_paranthesis = 0; to_take = TRUE; }  //constructor

	//take message onto lexical_analyser
	int take_input() { return lex.take_input(); }
	void destroy()	{ lex.destroy(); }

	//program interface routines
	float expression(); // takes care of add and sub

	};

/**** identifier_table class member functions: ****/

signed int identifier_table::index_of(const char* str)
		{
		int found = FALSE;

		for(int i=0; i<tot_var; i++)
			if (!strcmp(var_str[i],str))
				{ found = TRUE; break; }

		return ( (found) ? i : -1 );
		}

int identifier_table::new_entry(const char* str)
		{
		if (tot_var == MAX_VAR) error("Too Many Variables!!");
		strcpy(var_str[tot_var], str);
		value[tot_var] = 0.0;
		return (tot_var++);
		}

float identifier_table::value_of(const char* str)
		{
		signed num = index_of(str);
		if (num == -1) num = new_entry(str);
		return value[num];
		}

void identifier_table::set_value(const char* str, float val)
		{
		signed num = index_of(str);
		if (num == -1) num = new_entry(str);
		value[num] = val;
		}

/**** input_buffer class member functions ****/

int input_buffer::take_input()
	{
	cputs("\r\n Enter Here:  (^Z to end input, Esc to quit program) \r\n");

	char ch;
	int position = 0;
	int done = FALSE;
	int no_of_lines = 1;

   while (!done)
		switch(ch=getch())
			{
			case 8:								//backspace
				if (wherex() > 1)
					{
					putch(8);putch(32);putch(8);
					position--;
					}
				continue;
			case 13: 							//return key
				no_of_lines++;
				putch(13);  //return
				putch(10);  //linefeed
				inp_ptr[position++] = 13;
				continue;
			case 27:								//escape
				cputs("\r\n User Break.");
				exit(1);
			default:
				putch(inp_ptr[position++] = ch);
				if (ch==26) done = TRUE;		// ^Z i.e. end of input
				continue;
			}
	last_pos = position - 1;
	return no_of_lines;
	}

/**** lexical_analyser class member function definitions ****/

void lexical_analyser::get_number()
	{
	number_value = 0.0;

	char ch;
	int done = FALSE;
	int in_decimal = FALSE;
	float mult_fact = .1;

	while(!done)
		switch(ch = inp_buf.next_input())
			{
			case '.':
				if (in_decimal) error(" Two decimals in one number.");
				in_decimal = TRUE;
				break;

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (in_decimal)
					{
					number_value += ((ch-'0')*mult_fact);
					mult_fact *= .1;
					}
				else
					number_value = (number_value*10) + (ch-'0');
				break;
			default:
				inp_buf.take_back();
				done = TRUE;
			}
	}

void lexical_analyser::get_string()
	{
	char ch;
	int str_pos = 0;

	forever
		if ( isalpha( ch=inp_buf.next_input() ) )
			{
			if (str_pos==MAX_VAR_LEN)
				error ("Length of variable exceeds 32 characters");
			else
				string_value[str_pos++] = ch;
			}
		else
			{
			string_value[str_pos] = 0;
			inp_buf.take_back();
			break;
			}
	}

token_types lexical_analyser::get_token()
	{
	char ch;

	forever
		switch(ch = inp_buf.next_input())
			{
			case '0': case '1': case '2': case '3': case '4': case '5':
			case '6': case '7': case '8': case '9': case '.':
				inp_buf.take_back();
				get_number();
				return NUMBER;

			case '+': case '-': case '/': case '*':
			case '=': case '(': case ')':
				return (token_types)ch;

			case 26:		return END;
			case 13:    return EOS;

			default:
				if (isalpha(ch))
					{
					inp_buf.take_back();
					get_string();
					return STRING;
					}
				else
					if (ch!=32)	error("Invalid character encountered.");
			}
	}

/**** parser class member function defintions ****/

float parser::expression()
	{
	float t = term();

	switch(last_token)  //it has already been read in by term
		{
		case ADD:		return (t + expression());
		case SUBTRACT:	return (t - expression());

		case RP:		if (!in_paranthesis) error("Unexpected Right Paranthesis.");
		case EOS:
		case END:		return t;

		default: error("Expression Failure.");
		}
	}

float parser::term()
	{
	float t = primary();

	if (to_take)
		last_token = lex.get_token();
	else
		to_take = TRUE;

	switch(last_token)
		{
		case MULTIPLY:		return (t * term());
		case DIVIDE: 		return (t / term());

		case RP: if (!in_paranthesis) error("Unexpected Right Paranthesis.");
		case ADD:
		case SUBTRACT:
		case EOS:
		case END:         return t;


		default: error(" Term Failure");
		}
	}

float parser::primary()
	{
	last_token = lex.get_token();

	switch(last_token)
		{
		case NUMBER:	return lex.read_number();

		case SUBTRACT:	return (-primary());

		case STRING:
			{
			char temp[MAX_VAR_LEN+1];
			strcpy(temp, lex.read_string());
			last_token = lex.get_token();
			if (last_token == ASSIGN)
				{
				float t = expression();
				tab.set_value(temp, t);
				to_take = FALSE;
				return t;
				}
			else
				{
				to_take = FALSE;
				return tab.value_of(temp);
				}
			}

		case LP:
			{
			in_paranthesis++;
			float t = expression();
			if (last_token == RP)
				{ in_paranthesis--; return t; }
			else
				error("Missing Right Parenthesis");
			}

      case EOS:
		case END:		error("Primary expected at end of expr.");

		default: error(" Primary Failure.");
		}
	}

